@@ -1,13 +1,21 @@
 #!/usr/bin/python
-# Copyright (C) 2009-2011 - Curtis Hovey <sinzui.is at verizon.net>
+# Copyright (C) 2009-2012 - Curtis Hovey <sinzui.is at verizon.net>
 # This software is licensed under the MIT license (see the file COPYING).
 """Check for syntax and style problems."""
 
 from __future__ import with_statement
 
+
 __metaclass__ = type
 
-import compiler
+
+__all__ = [
+    'Reporter',
+    'UniversalChecker',
+    ]
+
+
+import _ast
 import htmlentitydefs
 import logging
 import mimetypes
@@ -31,32 +39,52 @@ from xml.parsers.expat import ErrorString, ExpatError
 from formatdoctest import DoctestReviewer
 
 import contrib.pep8 as pep8
-from contrib.pyflakes.checker import Checker
+from contrib.cssccc import CSSCodingConventionChecker
+from contrib.pyflakes.checker import Checker as PyFlakesChecker
 try:
     import cssutils
     HAS_CSSUTILS = True
 except ImportError:
     HAS_CSSUTILS = False
 
-# Javascript checking is available if spider money's js is available.
-js = subprocess.Popen(
-    ['which', 'js'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-js_exec, ignore = js.communicate()
-if js.returncode != 0:
-    JS = None
-else:
-    JS = js_exec.strip()
 
-__all__ = [
-    'Reporter',
-    'UniversalChecker',
-    ]
+def find_exec(names):
+    """Return the name of a GI enabled JS interpreter."""
+    if os.name != 'posix':
+        return None
+
+    for name in names:
+        js = subprocess.Popen(
+            ['which', name], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+        js_exec, ignore = js.communicate()
+        if js.returncode == 0:
+            return js_exec.strip()
+
+
+JS = find_exec(['gjs', 'seed'])
+
+
+DEFAULT_MAX_LENGTH = 80
+
+
+class PocketLintPyFlakesChecker(PyFlakesChecker):
+    '''PocketLint checker for pyflakes.
+
+    This is here to work around some of the pyflakes problems.
+    '''
+
+    def NAME(self, node):
+        '''Locate name. Ignore WindowsErrors.'''
+        if node.id == 'WindowsError':
+            return
+        return super(PocketLintPyFlakesChecker, self).NAME(node)
 
 
 class Reporter:
     """Common rules for checkers."""
     CONSOLE = object()
     FILE_LINES = object()
+    COLLECTOR = object()
 
     def __init__(self, report_type, treeview=None):
         self.report_type = report_type
@@ -65,18 +93,23 @@ class Reporter:
             self.treestore = self.file_lines_view.get_model()
         self.piter = None
         self._last_file_name = None
+        self.call_count = 0
+        self.error_only = False
+        self.messages = []
 
     def __call__(self, line_no, message, icon=None,
                  base_dir=None, file_name=None):
         """Report a message."""
+        if self.error_only and icon != 'error':
+            return
+        self.call_count += 1
+        args = (line_no, message, icon, base_dir, file_name)
         if self.report_type == self.FILE_LINES:
-            self._message_file_lines(
-                line_no, message, icon=icon,
-                base_dir=base_dir, file_name=file_name)
+            self._message_file_lines(*args)
+        elif self.report_type == self.COLLECTOR:
+            self._message_collector(*args)
         else:
-            self._message_console(
-                line_no, message, icon=icon,
-                base_dir=base_dir, file_name=file_name)
+            self._message_console(*args)
 
     def _message_console(self, line_no, message, icon=None,
                          base_dir=None, file_name=None):
@@ -101,6 +134,11 @@ class Reporter:
         self.treestore.append(
             self.piter, (file_name, icon, line_no, message, base_dir))
 
+    def _message_collector(self, line_no, message, icon=None,
+                         base_dir=None, file_name=None):
+        self._last_file_name = (base_dir, file_name)
+        self.messages.append((line_no, message))
+
 
 class Language:
     """Supported Language types."""
@@ -117,19 +155,23 @@ class Language:
     ZCML = object()
     DOCBOOK = object()
     LOG = object()
+    SQL = object()
 
     XML_LIKE = (XML, XSLT, HTML, ZPT, ZCML, DOCBOOK)
 
     mimetypes.add_type('application/x-zope-configuation', '.zcml')
     mimetypes.add_type('application/x-zope-page-template', '.pt')
     mimetypes.add_type('text/x-python-doctest', '.doctest')
+    mimetypes.add_type('text/x-twisted-application', '.tac')
     mimetypes.add_type('text/x-log', '.log')
     mime_type_language = {
         'text/x-python': PYTHON,
+        'text/x-twisted-application': PYTHON,
         'text/x-python-doctest': DOCTEST,
         'text/css': CSS,
         'text/html': HTML,
         'text/plain': TEXT,
+        'text/x-sql': SQL,
         'text/x-log': LOG,
         'application/javascript': JAVASCRIPT,
         'application/xml': XML,
@@ -176,13 +218,13 @@ class BaseChecker:
 
     The Decedent must provide self.file_name and self.base_dir
     """
-
-    def __init__(self, file_path, text, reporter=None):
+    def __init__(self, file_path, text, reporter=None, options=None):
         self.file_path = file_path
         self.base_dir = os.path.dirname(file_path)
         self.file_name = os.path.basename(file_path)
         self.text = text
         self.set_reporter(reporter=reporter)
+        self.options = options
 
     def set_reporter(self, reporter=None):
         """Set the reporter for messages."""
@@ -205,38 +247,50 @@ class BaseChecker:
         """Check the content."""
         raise NotImplementedError
 
+    @property
+    def check_length_filter(self):
+        '''Default filter used by default for checking line length.'''
+        if self.options:
+            return self.options.max_line_length
+        else:
+            return DEFAULT_MAX_LENGTH
+
 
 class UniversalChecker(BaseChecker):
     """Check and reformat doctests."""
 
-    def __init__(self, file_path, text=None, language=None, reporter=None):
+    def __init__(self, file_path, text,
+                 language=None, reporter=None, options=None):
         self.file_path = file_path
         self.base_dir = os.path.dirname(file_path)
         self.file_name = os.path.basename(file_path)
         self.text = text
         self.set_reporter(reporter=reporter)
         self.language = language
+        self.options = options
         self.file_lines_view = None
 
     def check(self):
         """Check the file syntax and style."""
         if self.language is Language.PYTHON:
-            PythonChecker(self.file_path, self.text, self._reporter).check()
+            checker_class = PythonChecker
         elif self.language is Language.DOCTEST:
-            DoctestReviewer(self.text, self.file_path, self._reporter).check()
+            checker_class = DoctestReviewer
         elif self.language is Language.CSS:
-            CSSChecker(self.file_path, self.text, self._reporter).check()
+            checker_class = CSSChecker
         elif self.language in Language.XML_LIKE:
-            XMLChecker(self.file_path, self.text, self._reporter).check()
+            checker_class = XMLChecker
         elif self.language is Language.JAVASCRIPT:
-            JavascriptChecker(
-                self.file_path, self.text, self._reporter).check()
+            checker_class = JavascriptChecker
         elif self.language is Language.LOG:
             # Log files are not source, but they are often in source code
             # trees.
             pass
         else:
-            AnyTextChecker(self.file_path, self.text, self._reporter).check()
+            checker_class = AnyTextChecker
+        checker = checker_class(
+            self.file_path, self.text, self._reporter, self.options)
+        checker.check()
 
 
 class AnyTextMixin:
@@ -249,9 +303,11 @@ class AnyTextMixin:
 
     def check_length(self, line_no, line):
         """Check the length of the line."""
-        if len(line) > 78:
+        max_length = self.check_length_filter
+        if len(line) > max_length:
             self.message(
-                line_no, 'Line exceeds 78 characters.', icon='info')
+                line_no, 'Line exceeds %s characters.' % max_length,
+                icon='info')
 
     def check_trailing_whitespace(self, line_no, line):
         """Check for the presence of trailing whitespace in the line."""
@@ -269,13 +325,6 @@ class AnyTextMixin:
 class AnyTextChecker(BaseChecker, AnyTextMixin):
     """Verify the text of the document."""
 
-    def __init__(self, file_path, text, reporter=None):
-        self.file_path = file_path
-        self.base_dir = os.path.dirname(file_path)
-        self.file_name = os.path.basename(file_path)
-        self.text = text
-        self.set_reporter(reporter=reporter)
-
     def check(self):
         """Call each line_method for each line in text."""
         for line_no, line in enumerate(self.text.splitlines()):
@@ -285,11 +334,25 @@ class AnyTextChecker(BaseChecker, AnyTextMixin):
             self.check_conflicts(line_no, line)
 
 
+class SQLChecker(BaseChecker, AnyTextMixin):
+    """Verify SQL style."""
+
+    def check(self):
+        """Call each line_method for each line in text."""
+        # Consider http://code.google.com/p/python-sqlparse/ to verify
+        # keywords and reformatting.
+        for line_no, line in enumerate(self.text.splitlines()):
+            line_no += 1
+            self.check_trailing_whitespace(line_no, line)
+            self.check_tab(line_no, line)
+            self.check_conflicts(line_no, line)
+
+
 class XMLChecker(BaseChecker, AnyTextMixin):
     """Check XML documents."""
 
     xml_decl_pattern = re.compile(r'<\?xml .*?\?>')
-    xhtml_doctype = """
+    xhtml_doctype = u"""
         <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
           "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
         """
@@ -301,7 +364,12 @@ class XMLChecker(BaseChecker, AnyTextMixin):
         parser = ElementTree.XMLParser()
         parser.entity.update(htmlentitydefs.entitydefs)
         offset = 0
-        text = self.text
+        try:
+            # This is a sanity check to work with the true text....
+            text = unicode(self.text.decode('utf-8').encode('utf-8'))
+        except UnicodeDecodeError:
+            # ...but this fallback is okay since this check is about markup.
+            text = self.text.decode('ascii', 'ignore').encode('utf-8')
         if text.find('<!DOCTYPE') == -1:
             # Expat requires a doctype to honour parser.entity.
             offset = 3
@@ -312,7 +380,7 @@ class XMLChecker(BaseChecker, AnyTextMixin):
                 start, end = match.span(0)
                 text = text[:start] + self.xhtml_doctype + text[end:]
         try:
-            root = ElementTree.parse(StringIO(text), parser)
+            ElementTree.parse(StringIO(text), parser)
         except (ExpatError, ParseError), error:
             if hasattr(error, 'code'):
                 error_message = ErrorString(error.code)
@@ -326,7 +394,7 @@ class XMLChecker(BaseChecker, AnyTextMixin):
                     error_lineno = 0
             else:
                 error_message, location = str(error).rsplit(':')
-                error_lineno = int(location.split(',')[0].split()[1])- offset
+                error_lineno = int(location.split(',')[0].split()[1]) - offset
             self.message(error_lineno, error_message, icon='error')
         self.check_text()
 
@@ -378,16 +446,27 @@ class CSSChecker(BaseChecker, AnyTextMixin):
 
     def check(self):
         """Check the syntax of the CSS code."""
-        if self.text == '' or not HAS_CSSUTILS:
+        if self.text == '':
+            return
+
+        self.check_cssutils()
+        self.check_text()
+        # CSS coding conventoins checks should go last since they rely
+        # on previous checks.
+        self.check_css_coding_conventions()
+
+    def check_cssutils(self):
+        """Check the CSS code by parsing it using CSSUtils module."""
+        if not HAS_CSSUTILS:
             return
         handler = CSSReporterHandler(self)
         log = logging.getLogger('pocket-lint')
         log.addHandler(handler)
+        log.propagate = False
         parser = cssutils.CSSParser(
             log=log, loglevel=logging.INFO, raiseExceptions=False)
         parser.parseString(self.text)
         log.removeHandler(handler)
-        self.check_text()
 
     def check_text(self):
         """Call each line_method for each line in text."""
@@ -398,36 +477,43 @@ class CSSChecker(BaseChecker, AnyTextMixin):
             self.check_conflicts(line_no, line)
             self.check_tab(line_no, line)
 
+    def check_css_coding_conventions(self):
+        """Check the input using CSS Coding Convention checker."""
+        CSSCodingConventionChecker(self.text, logger=self.message).check()
+
 
 class PythonChecker(BaseChecker, AnyTextMixin):
     """Check python source code."""
 
-    def __init__(self, file_path, text, reporter=None):
+    # This regex is taken from PEP 0263.
+    encoding_pattern = re.compile("coding[:=]\s*([-\w.]+)")
+
+    def __init__(self, file_path, text, reporter=None, options=None):
         super(PythonChecker, self).__init__(
-            file_path, text, reporter=reporter)
-        self.is_utf8 = False
+            file_path, text, reporter, options)
+        self.encoding = 'ascii'
 
     def check(self):
         """Check the syntax of the python code."""
         if self.text == '':
             return
+        self.check_text()
         self.check_flakes()
         self.check_pep8()
-        self.check_text()
 
     def check_flakes(self):
         """Check compilation and syntax."""
         try:
-            tree = compiler.parse(self.text)
+            tree = compile(
+                self.text, self.file_path, "exec", _ast.PyCF_ONLY_AST)
         except (SyntaxError, IndentationError), exc:
             line_no = exc.lineno or 0
-            offset = exc.offset or 0
             line = exc.text or ''
             explanation = 'Could not compile; %s' % exc.msg
             message = '%s: %s' % (explanation, line.strip())
             self.message(line_no, message, icon='error')
         else:
-            warnings = Checker(tree)
+            warnings = PocketLintPyFlakesChecker(tree, self.file_path)
             for warning in warnings.messages:
                 dummy, line_no, message = str(warning).split(':')
                 self.message(int(line_no), message.strip(), icon='error')
@@ -442,26 +528,32 @@ class PythonChecker(BaseChecker, AnyTextMixin):
                 self.message(line_no, message, icon='info')
 
             pep8.Checker.report_error = pep8_report_error
+            original_max_line_length = pep8.MAX_LINE_LENGTH
+            pep8.MAX_LINE_LENGTH = self.check_length_filter
             pep8.process_options([self.file_path])
             try:
                 pep8.Checker(self.file_path).check_all()
             except TokenError, er:
                 message, location = er.args
                 self.message(location[0], message, icon='error')
+            except IndentationError, er:
+                message, location = er.args
+                message = "%s: %s" % (message, location[3].strip())
+                self.message(location[1], message, icon='error')
         finally:
-            Checker.report_error = original_report_error
+            pep8.MAX_LINE_LENGTH = original_max_line_length
+            pep8.Checker.report_error = original_report_error
 
     def check_text(self):
         """Call each line_method for each line in text."""
         for line_no, line in enumerate(self.text.splitlines()):
             line_no += 1
-            if line_no in (1, 2) and '# -*- coding: utf-8 -*-' in line:
-                self.is_utf8 = True
+            if line_no in (1, 2):
+                match = self.encoding_pattern.search(line)
+                if match:
+                    self.encoding = match.group(1).lower()
             self.check_pdb(line_no, line)
-            self.check_length(line_no, line)
-            self.check_trailing_whitespace(line_no, line)
             self.check_conflicts(line_no, line)
-            self.check_tab(line_no, line)
             self.check_ascii(line_no, line)
 
     def check_pdb(self, line_no, line):
@@ -471,12 +563,20 @@ class PythonChecker(BaseChecker, AnyTextMixin):
             self.message(
                 line_no, 'Line contains a call to pdb.', icon='error')
 
+    @property
+    def check_length_filter(self):
+        # The pep8 lib counts from 0.
+        if self.options:
+            return self.options.max_line_length - 1
+        else:
+            return pep8.MAX_LINE_LENGTH
+
     def check_ascii(self, line_no, line):
         """Check that the line is ascii."""
-        if self.is_utf8:
+        if self.encoding != 'ascii':
             return
         try:
-            ascii_line = line.encode('ascii')
+            line.encode('ascii')
         except UnicodeEncodeError, error:
             self.message(
                 line_no, 'Non-ascii characer at position %s.' % error.end,
@@ -494,7 +594,7 @@ class JavascriptChecker(BaseChecker, AnyTextMixin):
         """Check the syntax of the javascript code."""
         if JS is None or self.text == '':
             return
-        args = [JS, '-f', self.FULLJSLINT, self.JSREPORTER, self.text]
+        args = [JS, self.JSREPORTER, self.FULLJSLINT, self.file_path]
         jslint = subprocess.Popen(
             args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
         issues, errors = jslint.communicate()
@@ -502,13 +602,23 @@ class JavascriptChecker(BaseChecker, AnyTextMixin):
         if issues:
             for issue in issues.splitlines():
                 line_no, char_no_, message = issue.split('::')
-                self.message(int(line_no), message, icon='error')
+                line_no = int(line_no)
+                line_no -= 1
+                self.message(line_no, message, icon='error')
         self.check_text()
 
+    def check_debugger(self, line_no, line):
+        """Check the length of the line."""
+        debugger_call = 'debugger;'
+        if debugger_call in line:
+            self.message(
+                line_no, 'Line contains a call to debugger.', icon='error')
+
     def check_text(self):
         """Call each line_method for each line in text."""
         for line_no, line in enumerate(self.text.splitlines()):
             line_no += 1
+            self.check_debugger(line_no, line)
             self.check_length(line_no, line)
             self.check_trailing_whitespace(line_no, line)
             self.check_conflicts(line_no, line)
@@ -517,19 +627,36 @@ class JavascriptChecker(BaseChecker, AnyTextMixin):
 
 def get_option_parser():
     """Return the option parser for this program."""
-    usage = "usage: %prog [options] arg1"
+    usage = "usage: %prog [options] file1 file2"
     parser = OptionParser(usage=usage)
     parser.add_option(
-        "-v", "--verbose", action="store_true", dest="verbose")
+        "-v", "--verbose", action="store_true", dest="verbose",
+        help="show errors and warngings.")
     parser.add_option(
-        "-q", "--quiet", action="store_false", dest="verbose")
-    parser.set_defaults(verbose=True)
+        "-q", "--quiet", action="store_false", dest="verbose",
+        help="Show errors only.")
+    parser.add_option(
+        "-f", "--format", dest="do_format", action="store_true",
+        help="Reformat the doctest.")
+    parser.add_option(
+        "-i", "--interactive", dest="is_interactive", action="store_true",
+        help="Approve each change.")
+    parser.add_option(
+        "-m", "--max-length", dest="max_line_length", type="int",
+        help="Set the max line length (default %s)" % DEFAULT_MAX_LENGTH)
+    parser.set_defaults(
+        verbose=True,
+        do_format=False,
+        is_interactive=False,
+        max_line_length=DEFAULT_MAX_LENGTH,
+        )
     return parser
 
 
-def check_sources(sources, reporter=None):
+def check_sources(sources, options, reporter=None):
     if reporter is None:
         reporter = Reporter(Reporter.CONSOLE)
+    reporter.call_count = 0
     for source in sources:
         file_path = os.path.normpath(source)
         if os.path.isdir(source) or not Language.is_editable(source):
@@ -537,9 +664,13 @@ def check_sources(sources, reporter=None):
         language = Language.get_language(file_path)
         with open(file_path) as file_:
             text = file_.read()
+        if language is Language.DOCTEST and options.do_format:
+            formatter = DoctestReviewer(text, file_path, reporter)
+            formatter.format_and_save(options.is_interactive)
         checker = UniversalChecker(
-            file_path, text=text, language=language, reporter=reporter)
+            file_path, text, language, reporter, options=options)
         checker.check()
+    return reporter.call_count
 
 
 def main(argv=None):
@@ -551,10 +682,9 @@ def main(argv=None):
     # Handle standard args.
     if len(sources) == 0:
         parser.error("Expected file paths.")
-    if options.verbose:
-        pass
     reporter = Reporter(Reporter.CONSOLE)
-    check_sources(sources, reporter)
+    reporter.error_only = not options.verbose
+    return check_sources(sources, options, reporter)
 
 
 if __name__ == '__main__':
